/**
 * Program      cosineWaveGenerator.cpp
 * Author       2021-05-19 Charles Geiser (https://www.dodeka.ch)
 * 
 * Purpose      Shows how to use the builtin cosine wave generator of the ESP32
 * 
 * Board        ESP32 DoIt DevKit V1
 *
 * Wiring       Oscilloscope on output DAC_1 (GPIO25) and DAC_2 (GPIO26)
 *
 * Remarks      The common output frequency of both DAC channels is given by the formula:
 * 
 *              freq = dig_clk_rtc_freq / (1 + RTC_CNTL_CK8M_DIV_SEL) * SENS_SAR_SW_FSTEP / 65536
 *              =================================================================================
 *              with
 *              dig_clk_rtc_freq = 8'000'000 Hz (assumed operating frequency, no need to know it exactly, see below)
 *              f0 = dig_clk_rtc_freq / 65536 = 8000000 / 65536 = 122.0703125
 *              we get
 *              freq = f0 * SENS_SAR_SW_FSTEP / (1 + RTC_CNTL_CK8M_DIV_SEL)
 *              freq = f0 * step / (1 + divi) 
 *              =============================
 *              RTC_CNTL_CK8M_DIV_SEL = 0..7
 *              SENS_SAR_SW_FSTEP     = 1..65535 (0x0001 .. 0xffff)
 * 
 *              If we set SENS_SAR_SW_FSTEP = 1 and RTC_CNTL_CK8M_DIV_SEL = 0 
 *              we are able to measure f0 directly.
 *              This referenc frequency is ouput on startup of the program.
 *              We measure this frequency with a oscilloscope or a counter and 
 *              set it in the code or on the command line of the program.
 *               
 *
 * References   https://www.espressif.com/sites/default/files/documentation/esp32_technical_reference_manual_en.pdf
 *              https://github.com/krzychb/dac-cosine
 *              https://www.esp32.com/viewtopic.php?t=10321
 */

#include <Arduino.h>
#include "CosineWaveGenerator.h"

#define CLR_LINE  "\r                                                                              \r"
#define dig_clk_rtc_freq  8000000.0             // 8 MHz assumed operating frequency
constexpr double f0 = dig_clk_rtc_freq / 65536; // = 122.0703125 resulting reference frequency, change it to your measured frequency on startup

typedef struct { const char key; const char *txt; void (&action)(); } MenuItem;

CosineWaveGenerator cwGen(f0);

// Forward declaration of the menu actions
void toggleChannel1();
void toggleChannel2();
void setScale1();
void setScale2();
void setFrequencyWithSetDivisor();
void setFrequencyWithSetStep();
void setFrequencyBestMatch();
void setDivisor();
void setStep();
void setReferenceFrequency();
void setMode1();
void setMode2();
void setOffset1();
void setOffset2();
void setTolerance();
void printCurrentValues();
void showMenu();

MenuItem menu[] = 
{
  { 'r', "[r] Set reference frequency f0",    setReferenceFrequency },
  { 'c', "[c] Toggle CHN_1 on/off",           toggleChannel1 },
  { 'C', "[C] Toggle CHN_2 on/off",           toggleChannel2 },
  { 'd', "[d] Set clock divisor 0..7",        setDivisor },
  { 's', "[s] Set step 1..65535",             setStep },
  { 'f', "[f] Set frequenz with set divisor", setFrequencyWithSetDivisor },
  { 'F', "[F] Set frequenz with set step",    setFrequencyWithSetStep },
  { 'b', "[b] Set frequency best match with set tolerance", setFrequencyBestMatch },
  { 't', "[t] Set tolerance for best match (1..999 °/oo)", setTolerance },
  { 'm', "[m] Set CHN_1 mode 0..3",           setMode1 },
  { 'M', "[M] Set CHN_2 mode 0..3",           setMode2 },
  { 'a', "[a] Set CHN_1 amplitude 0..3",      setScale1 },
  { 'A', "[A] Set CHN_2 amplitude 0..3",      setScale2 },  
  { 'o', "[o] Set CHN_1 offset 0..255",       setOffset1 },
  { 'O', "[o] Set CHN_2 offset 0..255",       setOffset2 },
  { 'p', "[p] Print current values",           printCurrentValues },
  { 'S', "[S] Show Menu",                     showMenu }
};
constexpr uint8_t nbrMenuItems = sizeof(menu) / sizeof(menu[0]);

/**
 * Executes the action assigned to the key
 */
void doMenu()
{
  char key = Serial.read();
  for (int i = 0; i < nbrMenuItems; i++)
  {
    if (key == menu[i].key)
    {
      menu[i].action();
      break;
    }
  }
}

/**
 * Switch on or off channel 1
 */
void toggleChannel1()
{
  cwGen.toggle(DAC_CHANNEL_1);
}

/**
 * Switch on or off channel 2
 */
void toggleChannel2()
{
  cwGen.toggle(DAC_CHANNEL_2);
}

/**
 * Set the reference frequency
 * After startup the generator runs with this frequency. 
 * It can be measured and entered with this function  
 * to get correct results for the further operations
 */
void setReferenceFrequency()
{
  double value = f0;
  Serial.printf(CLR_LINE);
  Serial.printf("Enter reference frequency: ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseFloat();
  }
  if (value < 100.0 || value > 150.0)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 100.0 .. 150.0 (Hz)");
      return;
    }
  Serial.printf(CLR_LINE);
  Serial.printf("Reference frequency set to f0 = %.2f", value);
  cwGen.setReferenceFrequency(value);
}

/**
 * Sets the frequency without changing the previously set divisor 
 */
void setFrequencyWithSetDivisor()
{
  double value = 0.0;
  int divi = cwGen.getClockDivisor();
  int step = cwGen.getFrequencyStep();

  Serial.printf(CLR_LINE);
  Serial.printf("Enter frequency: ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseFloat();
  }
  Serial.printf(CLR_LINE);
  if (value < 15.0 || value > 8000000.0)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 15 .. 8'000'000 (Hz)");
      cwGen.setFrequency(0, 1);  // set reference frequency
      return;
    }
  cwGen.setFrequencyWithDivisor(value, divi);
  Serial.printf("f_target set to %.2f with divisor set to %d and computed step %d, f_actual = %.2f ", value, divi, step, cwGen.getActualFrequency());
}

/**
 * Sets the frequency without changing the previously set frequency step 
 */
void setFrequencyWithSetStep()
{
  double value = 0.0;

  Serial.printf(CLR_LINE);
  Serial.printf("Enter frequency: ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseFloat();
  }
  Serial.printf(CLR_LINE);
  if (value < 15.0 || value > 8000000.0)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 15 .. 8'000'000 (Hz)");
      cwGen.setFrequency(0, 1);  // set reference frequency
      return;
    }
  int step = cwGen.getFrequencyStep();
  cwGen.setFrequencyWithStep(value, step);
  int divi = cwGen.getClockDivisor();
  Serial.printf("f_target set to %.2f with step set to %d and computed divisor %d, f_actual = %.2f ", value, step, divi, cwGen.getActualFrequency());
}

/**
 * Sets the divisor and the frequency step so that the desired 
 * frequency is set with the greatest achievable accuracy
 */
void setFrequencyBestMatch()
{
  double value = 0.0;

  Serial.printf(CLR_LINE);
  Serial.printf("Enter frequency: ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseFloat();
  }
  Serial.printf(CLR_LINE);
  if (value < 15.0 || value > 1000000.0)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 15 .. 1'000'000 (Hz)");
      cwGen.setFrequency(0, 1);  // set reference frequency
      return;
    }
  cwGen.setFrequency(value);
  //int step = cwGen.getFrequencyStep();
  //int divi = cwGen.getClockDivisor();
  //Serial.printf("f_target set to %.2f with step %d and divisor %d, f_actual = %.2f ", value, step, divi, cwGen.getActualFrequency());
}

/**
 * Set the divisor
 */
void setDivisor()
{
  int value = 0;
  Serial.printf(CLR_LINE);  
  Serial.printf("Enter divider: ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseInt();
  }
  Serial.printf(CLR_LINE);
  if (value < 0 || value > 7)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 0 .. 7");
      return;
    }
  cwGen.setClockDivisor(value);
  Serial.printf("Divisor set to: %d ", value);
}

/**
 * Set the frequency step
 */
void setStep()
{
  uint32_t value = 1;
  Serial.printf(CLR_LINE);  
  Serial.printf("Enter step: ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseInt();
  }
  Serial.printf(CLR_LINE);
  if (value < 1 || value > 65535)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 1 .. 65535");
      return;
    }
  cwGen.setFrequencyStep(value);
  Serial.printf("Step set to: %d ", value);
}

/**
 * Scale output amplitude of a DAC channel using two bit pattern
 *  0 full swing
 *  1 scale to 1/2
 *  2 scale to 1/4
 *  3 scale to 1/8
 */
void setScale(dac_channel_t chn)
{
  int value = 0;
  Serial.printf(CLR_LINE);  
  Serial.printf("Enter scale: ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseInt();
  }
  Serial.printf(CLR_LINE);
  if (value < 0 || value > 3)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 0 .. 3");
      return;
    }
  cwGen.setScale(chn, value);
  Serial.printf("Scale set to: %d ", value);
}

/**
 * Scale output amplitude of channel 1 
 */
void setScale1()
{
  setScale(DAC_CHANNEL_1);
}

/**
 * Scale output amplitude of channel 2 
 */
void setScale2()
{
  setScale(DAC_CHANNEL_2);
}

/**
 * Set mode of output channel
 */
void setMode(dac_channel_t chn)
{
  int value = 0;
  Serial.printf(CLR_LINE);  
  Serial.printf("Enter mode: ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseInt();
  }
  Serial.printf(CLR_LINE);
  if (value < 0 || value > 3)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 0 .. 3");
      return;
    }
  cwGen.setMode(chn, (CWmode)value);
  Serial.printf("Mode set to: %d ", value);
}

/**
 * Set mode of output channel 1
 */
void setMode1()
{
  setMode(DAC_CHANNEL_1);
}

/**
 * Set mode of output channel 2
 */
void setMode2()
{
  setMode(DAC_CHANNEL_2);
}

/**
 * Set offset of output channel
 */
void setOffset(dac_channel_t chn)
{
  int value = 0;
  Serial.printf(CLR_LINE);  
  Serial.printf("Enter offset: ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseInt();
  }
  Serial.printf(CLR_LINE);
  if (value < 0 || value > 255)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 0 .. 255");
      return;
    }
  cwGen.setOffset(chn, value);
  Serial.printf("Offset set to: %d ", value);
}

/**
 * Set offset of output channel 1
 */
void setOffset1()
{
  setOffset(DAC_CHANNEL_1);
}

/**
 * Set offset of output channel 2
 */
void setOffset2()
{
  setOffset(DAC_CHANNEL_2);
}

/**
 * Set allowed tolerance of output frequency
 */
void setTolerance()
{
  int value = 0;
  Serial.printf(CLR_LINE);  
  Serial.printf("Enter tolerance in per thousend (1..999): ");
  delay(2000);
  while (Serial.available())
  {
    value = Serial.parseInt();
  }
  Serial.printf(CLR_LINE);
  if (value < 1 || value > 999)
    {
      Serial.printf(CLR_LINE);
      Serial.printf("Value out of range, allowed: 1 .. 999");
      return;
    }
  cwGen.setToleranceForBestMatch(value);
  Serial.printf("Tolerance set to: %d ", value);
}

void printCurrentValues()
{
  cwGen.printCwgData();
}

/**
 * Display menu on monitor
 */
void showMenu()
{
  Serial.printf("\n---------------------------\n");
  Serial.printf("ESP32 Cosine Wave Generator\n");
  Serial.printf("---------------------------\n");
  for (int i = 0; i < nbrMenuItems; i++)
    Serial.printf("%s\n", menu[i].txt);

  Serial.printf("\nPress a key: ");
}

void setup() 
{
  Serial.begin(115200);
  showMenu();
  cwGen.enable(DAC_CHANNEL_1);
  cwGen.enable(DAC_CHANNEL_2);
}

void loop() 
{
  if(Serial.available())
  {
    doMenu();
  }
}